<link rel="stylesheet" href="./sly/css/app/widgets/container/style.css" />
<div
  {%
  if
  widget._flex_direction == 'vertical'
  %}
  class="container"
  {%
  else
  %}
      {%
      if
      widget._overflow == 'scroll'
      %}
      class="container scroll"
      {%
      elif
      widget._overflow == 'wrap'
      %}
      class="container wrap"
      {%
      else
      %}
      class="container"
      {%
      endif
      %}
  {%
  endif
  %}
  style="flex-direction: {{{widget._flex_direction}}}; gap: {{{widget._gap}}}px"
>
  {% for item in widget._widgets %} 
    {% if widget._fractions is none %}
      <div>{{{item}}}</div>
    {% else %}
      <div style="flex: {{{widget._fractions[loop.index0]}}}; display: flex; ">
        {{{item}}}
      </div>
    {% endif %} 
  {% endfor %}
</div>

